import 'package:flutter/material.dart';
import 'package:flutter/foundation.dart';
import 'iroute_root.dart';
import 'url_parser.dart';
import 'page_list.dart';

typedef RouteBuilder = Widget Function(
    BuildContext context, Map<String, dynamic> params);

class IRoute extends RouterDelegate<String>
    with ChangeNotifier, PopNavigatorRouterDelegateMixin<String> {
  GlobalKey<NavigatorState> get navigatorKey =>
      GlobalObjectKey<NavigatorState>(this);

  static IRoute of(BuildContext context) {
    final delegate = Router.of(context).routerDelegate;
    assert(delegate is IRoute, 'Delegate type must match');
    return delegate as IRoute;
  }

  late Map<String, RouteBuilder> routes;
  final String defaultUrl;
  final String unknowUrl;
  IRouteRoot? _root;
  bool get isRoot => _root == null;
  set root(value) {
    assert(value != null);
    this._root = value;
    _root!.nested = this;
    _root!.addListener(this.notifyListeners);
  }

  PageList? _pageList;
  set pageList(PageList value) {
    this._pageList = value;
  }

  //error:Null check operator used on a null value
  int get selectedPage {
    if (_pageList == null) {
      print('_pagelist=null,return 0');
      return 0;
    }
    return _pageList!.current;
  }

  set selectedPage(int value) {
    _pageList!.current = value;
    push(_pageList!.pages[_pageList!.current].currentUrl);
    //notifyListeners();
  }

  @override
  void dispose() {
    super.dispose();
    if (!isRoot) _root!.removeListener(this.notifyListeners);
  }

  IRoute(
      {required this.routes,
      this.defaultUrl = '/',
      this.unknowUrl = '/404',
      IRouteRoot? root})
      : _root = root ?? root {
    //
    if (routes.containsKey('_')) {
      addHome = (context, params) => routes['_']!(context, params);
    }
  }

  RouteBuilder? addHome;
  List<String> _routeStack = [];
  List<String> get routeStack => _routeStack;

  bool _canPop = true;
  bool get canPop {
    if (_canPop == false) return false;

    return _routeStack.length > 1;
  }

  set canPop(bool canPop) => _canPop = canPop;

  String? get currentConfiguration {
    if (isRoot) {
      var result = routeStack.isNotEmpty ? routeStack.last : null;
      return result;
    }
  }

  Widget _getWidget(context, url) {
    String pattern = UrlParser.getPattern(url);
    var params = UrlParser.getParam(url);
    if (!isRoot) {
      if (params.containsKey('page'))
        _pageList!.current = params['page'];
      else
        _pageList!.current = 0;

      if (_pageList!.pages[_pageList!.current].widget == null ||
          _pageList!.pages[_pageList!.current].currentUrl != url) {
        if (addHome != null)
          _pageList!.pages[_pageList!.current].widget = addHome!(context, {
            'child':
                ((context) => this.routes[pattern]!(context, params))(context)
          });
        else
          _pageList!.pages[_pageList!.current].widget =
              ((context) => this.routes[pattern]!(context, params))(context);
      }
      if (_pageList!.pages[_pageList!.current].currentUrl != url)
        _pageList!.pages[_pageList!.current].currentUrl = url;

      return _pageList!.pages[_pageList!.current].widget!;
    } else
      return ((context) => this.routes[pattern]!(context, params))(context);
  }

  Page buildPage(context, url) {
    if (url == this.defaultUrl)
      return MaterialPage(
          key: UniqueKey(),
          name: '/',
          child: this._getWidget(
              context, this.defaultUrl) //add argument like(id:1) here
          );
    if (url == this.unknowUrl)
      return MaterialPage(
          key: UniqueKey(), name: url, child: this._getWidget(context, url));
    return MaterialPage(
        key: UniqueKey(), name: url, child: _getWidget(context, url));
  }

  @override
  Widget build(BuildContext context) {
    if (this._routeStack.isEmpty) _routeStack.add(this.defaultUrl);
    List<Page<dynamic>> _pages = [];
    for (String path in routeStack) _pages.add(buildPage(context, path));

    return Navigator(
        key: navigatorKey,
        pages: _pages,
        onPopPage: (route, result) {
          if (!route.didPop(result)) {
            return false;
          }
          if (canPop) {
            pop();
          }
          return true;
        });
  }

  @override
  Future<void> setNewRoutePath(String path) async {
    if (_canPop == false) return;
    if (path == currentConfiguration) return;
    _routeStack = _setNewRouteHistory(_routeStack, path);
    if (!isRoot) {
      var params = UrlParser.getParam(path);
      if (params.containsKey('page'))
        _pageList!.current = params['page'];
      else
        _pageList!.current = 0;
      _root!.currentConfiguration = this._routeStack.last;
    } else
      notifyListeners();

    return;
  }

  @override
  Future<bool> popRoute() {
    print('Pop Route');
    return super.popRoute();
  }

  List<String> _setNewRouteHistory(List<String> routes, String newRoute) {
    List<String> pathsHolder = [];
    pathsHolder.addAll(routes);
    for (String path in routes) {
      if (path == newRoute) {
        int index = routes.indexOf(path);
        int count = routes.length;
        for (var i = index; i < count - 1; i++) {
          pathsHolder.removeLast();
        }
        return pathsHolder;
      }
    }
    pathsHolder.add(newRoute);
    return pathsHolder;
  }

  void push(String path) {
    if (path == _routeStack.last) return;
    String url = path;

    if (_pageList != null) {
      if (!path.contains(':ipage') && path != '/') {
        url = url + ':ipage=${_pageList!.current}';
      }
    }

    _routeStack.add(url);
    if (!isRoot)
      _root!.currentConfiguration = this._routeStack.last;
    else
      notifyListeners();
  }

  void pop() {
    if (!canPop) return;
    _routeStack.removeLast();
    if (!isRoot) {
      var params = UrlParser.getParam(_routeStack.last);
      if (params.containsKey('page'))
        _pageList!.current = params['page'];
      else
        _pageList!.current = 0;
      _root!.currentConfiguration = this._routeStack.last;
    } else
      notifyListeners();
  }
}
